# University of Michigan Shibboleth Authentication for `python-social-auth`

This library contains backends and instructions for integrating python applications into
[University of Michigan's Shibboleth single sign-on](https://its.umich.edu/accounts-access/shibboleth)
infrastructure.

## Installation

The package is not currently available in a publicly accessible package repository.
However, you can still install this package as an external git dependency from Github
using `pip` or similar tools.

### Setting up an SSH Key with Github

Follow the [Github instructions](https://docs.github.com/en/authentication/connecting-to-github-with-ssh) for configuring an SSH key for your account

### Use directly in your active environment

Using `pip`, you can install this package directly from Github like the following:

```shell
pip install git+ssh://git@github.com/academic-innovation/social-auth-umich.git@main#egg=social-auth-umich
```

### In Requirements files

These same format can be used within a `requirements.txt` file, which can be useful for containerized builds and deploys. E.g.

```
social-auth-umich @ git+ssh://git@github.com/academic-innovation/social-auth-umich.git@main
```

### Package Extras

There are three package extra definitions that can be used by the consuming project.

<dl>
    <dt>django</dt>
    <dd>
        Adds dependencies for Django >= 3.2, and an associated social-auth-app-django.
    </dd>
    <dt>saml</dt>
    <dd>Adds dependency on <code>python3-saml</code></dd>
    <dt>openidconnect</dt>
    <dd>Adds JWT library dependencies</dd>
</dl>

These are not crucial, but can be useful shortcuts for installing this library for your
project. For example, if this was being used with OIDC on a django project, you might
set your requirements to be `social-auth-umich[django,openidconnect]`.

## Protocol Options

There are
[multiple options](https://its.umich.edu/accounts-access/shibboleth/protocol-options)
for authenticating users with Shibboleth at UM. This supports two of those methods:

- Security Assertion Markup Language (SAML)
- OpenID Connect (OIDC)

SAML is supported directly by `python-social-auth`, and it's not strictly necessary to
install this library to use it.

## Registering an app with ITS

Regardless of which protocol you choose, you must register your app by filling out the
[Shibboleth Configuration Request Form](https://its.umich.edu/accounts-access/shibboleth/configuration-request-form).
The protocol you choose changes the information you need to give on that form. See the
details for each protocol to see what information is necessary.

## Using OIDC

A Social Auth backend for OIDC is available as: `social_auth_umich.backends.UMichOIDC`.

### Installation

To use as a dependency in your local project, you will either want to install this
package with its `openidconnect` extra, or `social-auth-core[openidconnect]`.

### Information required for registration

OIDC only requires two pieces of information: the Main URL of the application you are
registering (e.g. https://my-awesome-app.umich.edu/), and the redirect URI after a
successful authentication, which depends on your configuration below.

### Configuration

There are three (two required, one optional) settings to configure for this backend.

<dl>
    <dt>UMICH_OIDC_KEY</dt>
    <dd>The key given by UM IAM after registering the application</dd>
    <dt>UMICH_OIDC_SECRET</dt>
    <dd>The secret given by UM IAM after registering the application</dd>
    <dt>UMICH_OIDC_ENDPOINT</dt>
    <dd>
        A protocol and hostname for the target OIDC endpoint, without a trailing slash.
        By default, this is the production Shibboleth instance. For testing purposes,
        you may point this to the staging instance (currently
        "https://shib-idp-staging.dsc.umich.edu")
    </dd>
</dl>

### Django example

Configure the project as described in the [`python-social-auth` documentation](https://python-social-auth.readthedocs.io/en/latest/configuration/django.html).

Add `social_auth_umich.backends.UMichOIDC` to the `AUTHENTICATION_BACKENDS`:

```python
AUTHENTICATION_BACKENDS = (
    "social_auth_umich.backends.UMichOIDC",
    "django.contrib.auth.backends.ModelBackend",
)
```

For convience, you may wish to add `social_auth_umich.django` to the list of
`INSTALLED_APPS`. By doing this, you can use the `umich_oidc_redirect_path` management
command to get the path when registering the app. E.g.

```shell
$ python manage.py umich_oidc_redirect_path
/social/complete/umich-oidc/
$
```

Set the necessary configuration settings as defined above:

```python
SOCIAL_AUTH_UMICH_OIDC_KEY = "<key>"
SOCIAL_AUTH_UMICH_OIDC_SECRET = "<secret>"
```

In your templates, you can link to the backend URL using a link similar to the following:

```html
<a href="{% url 'social:begin' 'umich-oidc' %}">Sign in</a>
```

## Using SAML

> SAML is currently untested. If there are inaccuracies, please create a Merge Request.

The existing `social_auth.backends.saml.SAMLAuth` is enough to support SAML authentication
with UM Shibboleth.

### Installation

Ensure that the `python-social-auth` SAML dependencies are installed. The backend
requires `python3-saml`, which in turn requires a number of system libraries to
function, namely `libxml2` and `xmlsec`. Ensure that these are available on your system
before installing the python libraries.

#### Using Debian-based Linux distributions

```shell
$ apt-get install libxml2-dev libxmlsec1-dev libxmlsec1-openssl
```

#### Using Red Hat-based Linux distributions

> TODO

#### Using Homebrew (on macOS)

If you have [Homebrew](https://brew.sh/) installed already, you can use it by executing
the following command:

```shell
$ brew install libxml2 libxmlsec1 pkg-config
```

### Creating certificates

To establish trust with the Shibboleth IdP, your application must have its own set of
X.509 certificates to signing and validating communications. One way to create those
certificates is using `openssl` to generate a pair of self-signed certificates, e.g.:

```shell
$ openssl req -x509 -newkey rsa:4096 -keyout certs/key.pem -out certs/cert.pem -nodes -days 1096
```

The `cert.pem` contains the public key for the service provider. This is what you will
share with ITS. The `key.pem` contains the private key for this service provider.
**DO NOT SHARE THIS KEY**

### Information required for registration

ITS will request some information about the site's Entity ID (a URL to identify the
service, defined in your project's settings), the site's domain/hostname, and a URL to
an XML Metadata file. `python-social-auth` can generate this for us, as shown in the
Django example below. When asked for the "Connection method", you can choose "Other".

### Configuration

SAML configuration requires a number of settings, and it's best to consult the
[`python-social-auth` SAML documentation](https://python-social-auth.readthedocs.io/en/latest/backends/saml.html#required-configuration).
Some of the configuration matter is available in `social_auth_umich.saml`, which should
simplify some of the contents. When setting up the list of IdPs for your site, you can
use `social_auth_umich.saml.UMICH_IDP` (or
`social_auth_umich.saml.UMICH_STAGING_IDP`, for attaching to UM’s staging IdP) as
the basis for your `SOCIAL_AUTH_SAML_ENABLED_IDPS`. For example:

```python
from social_auth_umich.saml import UMICH_IDP

SOCIAL_AUTH_SAML_ENABLED_IDPS = {
    "umich": UMICH_IDP,
}
```

### SAML Attributes

This library includes some useful definitions for handling the parsing and mapping of
SAML attributes in the authentication response. The
`social_auth_umich.saml.SAMLAttributes` namespace could be helpful when customizing your
[`python-social-auth` authentication pipeline](https://python-social-auth.readthedocs.io/en/latest/pipeline.html#authentication-pipeline). These are based on the Shibboleth
attributes [documented by ITS](https://documentation.its.umich.edu/node/262/).

During the pipeline execution, the `details` dictionary is populated with a set of
attributes defined in the IdP configuration. The default IdP definitions include
attribute mapping for the `details` dictionary as follows:

<table>
  <thead>
    <tr>
      <th><code>user_details</code> field</th>
      <th>SAML attribute name</th>
      <th><code>SAMLAttributes</code> name</th>
    </tr>
  </thead>
  <tbody>
    <tr><td>full_name</td><td>displayName</td><td>MCOMMUNITY_NAME</td></tr>
    <tr><td>first_name</td><td>givenName</td><td>FIRST_NAME</td></tr>
    <tr><td>last_name</td><td>sn</td><td>LAST_NAME</td></tr>
    <tr><td>username</td><td>eduPersonPrincipalName</td><td>PRINCIPAL_NAME</td></tr>
    <tr><td>email</td><td>mail</td><td>EMAIL</td></tr>
  </tbody>
</table>

In addition, the `python-social-auth` "User ID" (`attr_permanent_user_id`) for the
response is mapped to `SAMLAttributes.PRINCIPAL_NAME`.

You can use the `*_BASE` dictionaries and mix-in the custom attribute mapping you need.
For instance, if you wanted to use the user's uniqname for their username (and you had
the `uid` SAML attribute granted):

```python
SOCIAL_AUTH_SAML_ENABLED_IDPS = {
    "umich": {
        **UMICH_IDP_BASE,
        "attr_username": SAMLAttributes.UNIQNAME,
    },
}
# Or, if you're targeting Python 3.9 or later, you can use Dictionary Union operations
SOCIAL_AUTH_SAML_ENABLED_IDPS = {
    "umich": UMICH_IDP_BASE | {
        "attr_username": SAMLAttributes.UNIQNAME,
    },
}
```

If you want to pull any SAML attribute values out of the SAML response during your own
pipeline stages, you can use `SAMLAttributes` to pull the information you need.

```python

def user_university_affiliation(backend, user, response, *args, **kwargs):
    if backend.name == "saml" and response["idp_name"] == "umich":
        um_affiliation = response["attributes"].get(SAMLAttributes.AFFILIATION, [])
        if 'student' in um_affiliation:
            ...
```

If there are values you want to save for later, you can use the `extra_data` feature in
`python-social-auth` to automatically store them.  You can do so by defining the setting
as mentioned in the
[`python-social-auth` SAML docs](https://python-social-auth.readthedocs.io/en/latest/backends/saml.html#advanced-settings),
e.g.:

```python
SOCIAL_AUTH_SAML_EXTRA_DATA = [
    (SAMLAttributes.AFFILIATION, "affiliation"),
    (SAMLAttributes.EMPLOYEE_NUMBER, "UMID"),
]
```

> NOTE: the `extra_data` mapping above is for all SAML backends, not UM specifically.

### Django Example

Configure the project as described in the [`python-social-auth` documentation](https://python-social-auth.readthedocs.io/en/latest/configuration/django.html).

Add `social_auth.backends.saml.SAMLAuth` to the `AUTHENTICATION_BACKENDS`:

```python
AUTHENTICATION_BACKENDS = (
    "social_auth.backends.saml.SAMLAuth",
    "django.contrib.auth.backends.ModelBackend",
)
```

Set the required `SOCIAL_AUTH_SAML_*` configuration values, as described in the
configuration section above.

```python
SOCIAL_AUTH_SAML_SP_ENTITY_ID = "https://my-awesome-app.umich.edu/shibboleth"
SOCIAL_AUTH_SAML_SP_PUBLIC_CERT = "<contents of cert.pem>"
# or read the contents of a file SOCIAL_AUTH_SAML_SP_PUBLIC_CERT = open("./certs/cert.pem", 'r').read()
# or read the contents of an env var SOCIAL_AUTH_SAML_SP_PUBLIC_CERT = os.environ['SAML_PUBLIC_KEY']
SOCIAL_AUTH_SAML_SP_PRIVATE_KEY = "<contents of key.pem>"
# or alternatives such as the above
SOCIAL_AUTH_SAML_ORG_INFO = {
    "en-US": {
        "name": "my-awesome-app",
        "displayname": "My Awesome App",
        "url": "https://my-awesome-app.umich.edu/",
    },
}
SOCIAL_AUTH_SAML_SUPPORT_CONTACT = {
    "givenName": "Gerald R. Ford, Jr.",
    "emailAddress": "prez38@umich.edu",
}
SOCIAL_AUTH_SAML_TECHNICAL_CONTACT = {
    "givenName": "James Earl Jones",
    "emailAddress": "darth-vader@umich.edu",
}
SOCIAL_AUTH_ENABLED_IDPS = {
    "umich": UMICH_IDP,
}
```

To help with our metadata generation, you can mount a view in
`social_auth_umich.django.views` to your urls.

In `urls.py`:

```
from django.urls import re_path
from social_auth_umich.django.views import saml_metadata_view


urlpatterns = [
    path('', include("social_django.urls", namespace="social")),
    re_path(r"^metadata\.xml$", saml_metadata_view),
    ...
]
```

You can then give "https://my-awesome-app.umich.edu/metadata.xml" as the "Metadata URL"
when registering your Service with ITS.

For convenience, there's also a `saml_begin_url` function in
`social_auth_umich.django.utils`, that makes generating a hyperlink within your project
simple. E.g.

```python
from django.shortcuts import render
from social_auth_umich.django.utils import saml_begin_url


def my_fancy_view(request):
    context = {
        "login_with_umich_url": saml_begin_url(),
    }
    return render(request, "fancy.html", context)

# Or, with a Class-based View
class MyFancyView(TemplateView):
    template_name = "fancy.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["login_with_umich_url"] = saml_begin_url()
        return context
```

In `fancy.html`:
```html
<a href="{{ login_with_umich_url }}">Log in with U-M</a>
```

## Development

This package is managed with [poetry](https://python-poetry.org/). All python assets
are formatted with [black](https://black.readthedocs.io/en/stable/) and checked with
[flake8](https://flake8.pycqa.org/). Python 3.6 or greater is required.
